@@ -1,7 +1,8 @@
+from copy import deepcopy
+
 import matplotlib.pyplot as plt
 import numpy as np
 import pandas as pd
-from copy import deepcopy
 
 from .ts_comparison import SeriesComparison, SeriesComparisonRelative
 from .ts_utils import unique_nans_in_series
@@ -90,13 +91,13 @@ class Detector:
             dtype = series.dtypes
             name = series.name
         elif isinstance(series, pd.DataFrame):
-            dtype = series.dtypes[0]
+            dtype = series.dtypes.values[0]
             name = series.columns[0]
         else:
             raise TypeError(
                 "Series must be pandas.Series or pandas.DataFrame!")
         # check dtype (of first col)
-        if dtype != float:
+        if not dtype in [float, np.float32]:
             raise TypeError("Series (or first column of DataFrame) must "
                             "have dtype float!")
         return name
@@ -415,8 +416,7 @@ class Detector:
         df : pandas.DataFrame
             results with flagged values set to NaN per applied rule.
         """
-        df = pd.concat([s for s in self.results.values()],
-                       axis=1)
+        df = pd.concat(self.results.values(), axis=1)
         df.columns = ["base series"] + list(self.ruleset.rules.keys())
         return df
 
@@ -458,6 +458,49 @@ class Detector:
         df.columns = list(self.ruleset.rules.keys())
         return df
 
+    def get_corrections_comparison(self, truth=None):
+
+        if truth is None and self.truth is not None:
+            truth = self.truth
+        else:
+            raise ValueError("Supply a time series for 'truth'!")
+
+        comments_traval = self.get_comment_series()
+        comments_traval.name = "traval_comment"
+
+        mask_truth_corrections = truth.iloc[:, 0].isna()
+        comments_truth = truth.loc[mask_truth_corrections]
+
+        k = list(self.comparisons.keys())[-1]
+        comparison = self.comparisons[k].comparison_series()
+        translate = {
+            -1: "Value modified",
+            0: "Flagged in both",
+            1: "Only flagged in 'truth' series",
+            2: "Only flagged in 'traval' series",
+            -9999: "NaN in both"
+        }
+        comparison = comparison.apply(lambda v: translate[v])
+        comparison.name = "comparison_label"
+
+        raw_index = (comments_traval.index
+                     .union(comments_truth.index))
+
+        truth.columns = ["truth_series", "truth_comment"]
+
+        traval_series = self.get_final_result()
+        traval_series.name = "traval_series"
+
+        df = pd.concat([
+            self.series.loc[raw_index.intersection(self.series.index)],
+            traval_series.loc[raw_index.intersection(traval_series.index)],
+            comments_traval,
+            truth.loc[raw_index.intersection(truth.index)],
+            comparison.loc[raw_index.intersection(comparison.index)]
+        ], axis=1)
+
+        return df
+
     def plot_overview(self, mark_suspects=True, **kwargs):
         """Plot timeseries with flagged values per applied rule.
 
